/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the Source EULA. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/

import * as azdata from 'azdata';
import { Deferred } from '../../common/promise';
import * as loc from '../../localizedConstants';
import { cssStyles } from '../../constants';
import { InitializingComponent } from '../components/initializingComponent';
import { PostgresModel } from '../../models/postgresModel';

export const validExtensions = ['citus', 'pgaudit', 'pgautofailover', 'pg_cron', 'pg_partman', 'plv8', 'postgis', 'postgis_raster', 'postgis_sfcgal', 'postgis_tiger_geocoder', 'tdigest'];

export class AddPGExtensionsDialog extends InitializingComponent {
	protected modelBuilder!: azdata.ModelBuilder;

	protected extensionsListInputBox!: azdata.InputBoxComponent;

	protected _completionPromise = new Deferred<string | undefined>();

	constructor(protected _model: PostgresModel) {
		super();
	}

	public showDialog(dialogTitle: string): azdata.window.Dialog {
		const dialog = azdata.window.createModelViewDialog(dialogTitle);
		dialog.cancelButton.onClick(() => this.handleCancel());
		dialog.registerContent(async view => {
			this.modelBuilder = view.modelBuilder;

			const info = this.modelBuilder.text().withProps({
				value: loc.extensionsAddFunction(validExtensions.join(', ')),
				CSSStyles: { ...cssStyles.text, 'margin-block-start': '0px', 'margin-block-end': '0px' }
			}).component();

			const link = this.modelBuilder.hyperlink().withProps({
				label: loc.extensionsLearnMore,
				url: 'https://docs.microsoft.com/azure/azure-arc/data/using-extensions-in-postgresql-hyperscale-server-group',
			}).component();

			const infoAndLink = this.modelBuilder.flexContainer().withLayout({ flexWrap: 'wrap' }).component();
			infoAndLink.addItem(info, { CSSStyles: { 'margin-right': '5px' } });
			infoAndLink.addItem(link);

			this.extensionsListInputBox = this.modelBuilder.inputBox()
				.withProps({
					value: '',
					ariaLabel: loc.extensionsAddList,
					enabled: true,
					validationErrorMessage: loc.extensionsAddErrorrMessage(validExtensions.join(','))
				}).withValidation((component) => {
					if (!component.value) {
						return true;
					}

					let newExtensions = component.value.split(',');
					return newExtensions.every(e => validExtensions.includes(e));
				}).component();

			let formModel = this.modelBuilder.formContainer()
				.withFormItems([{
					components: [
						{
							component: infoAndLink
						},
						{
							component: this.extensionsListInputBox,
							title: loc.extensionsAddList,
							required: true
						}
					],
					title: ''
				}]).withLayout({ width: '100%' }).component();
			await view.initializeModel(formModel);
			this.extensionsListInputBox.focus();
			this.initialized = true;
		});

		dialog.registerCloseValidator(async () => await this.validate());
		dialog.okButton.label = loc.loadExtensions;
		dialog.cancelButton.label = loc.cancel;
		azdata.window.openDialog(dialog);
		return dialog;
	}

	public async validate(): Promise<boolean> {
		this._completionPromise.resolve(this.extensionsListInputBox.value);
		return true;
	}

	private handleCancel(): void {
		this._completionPromise.resolve(undefined);
	}

	public waitForClose(): Promise<string | undefined> {
		return this._completionPromise.promise;
	}
}
